package java.lang.invoke;

import java.util.ArrayList;
import sun.reflect.CallerSensitive;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.lang.reflect.Proxy;
import sun.invoke.WrapperInstance;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationHandler;
import sun.reflect.misc.ReflectUtil;
import sun.reflect.Reflection;
import java.lang.reflect.Modifier;

public class MethodHandleProxies
{
    private MethodHandleProxies() {
    }
    
    @CallerSensitive
    public static <T> T asInterfaceInstance(final Class<T> clazz, final MethodHandle methodHandle) {
        if (!clazz.isInterface() || !Modifier.isPublic(clazz.getModifiers())) {
            throw MethodHandleStatics.newIllegalArgumentException("not a public interface", clazz.getName());
        }
        if (System.getSecurityManager() != null) {
            Class<Object> callerClass = Reflection.getCallerClass();
            ClassLoader ccl = (callerClass != null) ? callerClass.getClassLoader() : null;
            ReflectUtil.checkProxyPackageAccess(ccl, new Class[] { clazz });
            methodHandle = ((ccl != null) ? bindCaller(methodHandle, callerClass) : methodHandle);
        }
        else {
        }
        final ClassLoader classLoader = clazz.getClassLoader();
        if (classLoader == null) {
            ClassLoader contextClassLoader = Thread.currentThread().getContextClassLoader();
            classLoader = ((contextClassLoader != null) ? contextClassLoader : ClassLoader.getSystemClassLoader());
        }
        final Method[] singleNameMethods = getSingleNameMethods(clazz);
        if (singleNameMethods == null) {
            throw MethodHandleStatics.newIllegalArgumentException("not a single-method interface", clazz.getName());
        }
        final MethodHandle[] array = new MethodHandle[singleNameMethods.length];
        for (int i = 0; i < singleNameMethods.length; ++i) {
            Method method = singleNameMethods[i];
            MethodType methodType = MethodType.methodType(method.getReturnType(), method.getParameterTypes());
            MethodHandle type = methodHandle.asType(methodType);
            array[i] = type.asType(type.type().changeReturnType(Object.class)).asSpreader(Object[].class, methodType.parameterCount());
        }
        final Object h = new InvocationHandler() {
            private Object getArg(String s) {
                if (s == "getWrapperInstanceTarget") {
                    return methodHandle;
                }
                if (s == "getWrapperInstanceType") {
                    return clazz;
                }
                throw new AssertionError();
            }
            
            @Override
            public Object invoke(Object o, Method obj, Object[] array) throws Throwable {
                int i = 0;
                while (i < singleNameMethods.length) {
                    if (obj.equals(singleNameMethods[i])) {
                        return array[i].invokeExact(array);
                    }
                    ++i;
                }
                if (obj.getDeclaringClass() == WrapperInstance.class) {
                    return this.getArg(obj.getName());
                }
                if (MethodHandleProxies.access$000(obj)) {
                    return MethodHandleProxies.access$100(o, obj, array);
                }
                throw MethodHandleStatics.newInternalError("bad proxy method: " + obj);
            }
        };
        Object obj;
        if (System.getSecurityManager() != null) {
            obj = AccessController.doPrivileged(new PrivilegedAction<Object>() {
                @Override
                public Object run() {
                    return Proxy.newProxyInstance(classLoader, new Class[] { clazz, WrapperInstance.class }, h);
                }
            });
        }
        else {
            obj = Proxy.newProxyInstance(classLoader, new Class[] { clazz, WrapperInstance.class }, h);
        }
        return clazz.cast(obj);
    }
    
    private static MethodHandle bindCaller(MethodHandle methodHandle, Class<?> clazz) {
        MethodHandle bindCaller = MethodHandleImpl.bindCaller(methodHandle, clazz);
        if (methodHandle.isVarargsCollector()) {
            MethodType type = bindCaller.type();
            return bindCaller.asVarargsCollector(type.parameterType(type.parameterCount() - 1));
        }
        return bindCaller;
    }
    
    public static boolean isWrapperInstance(Object o) {
        return o instanceof WrapperInstance;
    }
    
    private static WrapperInstance asWrapperInstance(Object o) {
        try {
            if (o != null) {
                return (WrapperInstance)o;
            }
        }
        catch (ClassCastException ex) {}
        throw MethodHandleStatics.newIllegalArgumentException("not a wrapper instance");
    }
    
    public static MethodHandle wrapperInstanceTarget(Object o) {
        return asWrapperInstance(o).getWrapperInstanceTarget();
    }
    
    public static Class<?> wrapperInstanceType(Object o) {
        return asWrapperInstance(o).getWrapperInstanceType();
    }
    
    private static boolean isObjectMethod(Method method) {
        String name = method.getName();
        int n = -1;
        switch (name.hashCode()) {
            case -1776922004: {
                if (name.equals("toString")) {
                    n = 0;
                    break;
                }
                break;
            }
            case 147696667: {
                if (name.equals("hashCode")) {
                    n = 1;
                    break;
                }
                break;
            }
            case -1295482945: {
                if (name.equals("equals")) {
                    n = 2;
                    break;
                }
                break;
            }
        }
        switch (n) {
            case 0: {
                return method.getReturnType() == String.class && method.getParameterTypes().length == 0;
            }
            case 1: {
                return method.getReturnType() == Integer.TYPE && method.getParameterTypes().length == 0;
            }
            case 2: {
                return method.getReturnType() == Boolean.TYPE && method.getParameterTypes().length == 1 && method.getParameterTypes()[0] == Object.class;
            }
            default: {
                return false;
            }
        }
    }
    
    private static Object callObjectMethod(Object o, Method detailMessage, Object[] array) {
        assert !(!isObjectMethod(detailMessage)) : detailMessage;
        String name = detailMessage.getName();
        int n = -1;
        switch (name.hashCode()) {
            case -1776922004: {
                if (name.equals("toString")) {
                    n = 0;
                    break;
                }
                break;
            }
            case 147696667: {
                if (name.equals("hashCode")) {
                    n = 1;
                    break;
                }
                break;
            }
            case -1295482945: {
                if (name.equals("equals")) {
                    n = 2;
                    break;
                }
                break;
            }
        }
        switch (n) {
            case 0: {
                return o.getClass().getName() + "@" + Integer.toHexString(o.hashCode());
            }
            case 1: {
                return Integer.valueOf(System.identityHashCode(o));
            }
            case 2: {
                return Boolean.valueOf(o == array[0]);
            }
            default: {
                return null;
            }
        }
    }
    
    private static Method[] getSingleNameMethods(Class<?> clazz) {
        ArrayList<Method> list = new ArrayList();
        String s = null;
        for (Method e : clazz.getMethods()) {
            if (!isObjectMethod(e) && Modifier.isAbstract(e.getModifiers())) {
                String name = e.getName();
                if (s == null) {
                    s = name;
                }
                else if (!s.equals(name)) {
                    return null;
                }
                list.add(e);
            }
        }
        if (s == null) {
            return null;
        }
        return (Method[])list.toArray(new Method[list.size()]);
    }
    
    static boolean access$000(Method method) {
        return isObjectMethod(method);
    }
    
    static Object access$100(Object o, Method method, Object[] array) {
        return callObjectMethod(o, method, array);
    }
}
